package html4k.stream

import html4k.*
import html4k.Entities.*
import html4k.consumers.delayed
import org.w3c.dom.events.Event

class HTMLStreamBuilder<O : Appendable>(val out : O, val prettyPrint : Boolean) : TagConsumer<O> {
    private var level = 0
    private var ln = true

    override fun onTagStart(tag: Tag) {
        indent()
        level++

        out.append("<")
        out.append(tag.tagName)

        if (tag.attributes.isNotEmpty()) {
            tag.attributes.entrySet().map { e ->
                if (!e.getKey().isValidXmlAttributeName()) {
                    throw IllegalArgumentException("Tag ${tag.tagName} has invalid attribute name ${e.getKey()}")
                }
                StringBuilder {
                    append(e.getKey())
                    append("=\"")
                    escapeAppend(e.getValue())
                    append("\"")
                }
            }.joinTo(out, " ", " ")
        }

        out.append(">")
    }

    override fun onTagAttributeChange(tag : Tag, attribute: String, value: String) {
        throw UnsupportedOperationException("tag attribute can't be changed as it was already written to the stream. Use with DelayedConsumer to be able to modify attributes")
    }

    override fun onTagEvent(tag: Tag, event: String, value: (Event) -> Unit) {
        throw UnsupportedOperationException("you can't assign lambda event handler when building text")
    }

    override fun onTagEnd(tag: Tag) {
        level--
        if (prettyPrint && ln) {
            indent()
        }
        out.append("</")
        out.append(tag.tagName)
        out.append(">")

        appenln()
    }

    override fun onTagContent(content: CharSequence) {
        out.escapeAppend(content)
    }

    override fun onTagContentEntity(entity: Entities) {
        out.append(entity.text)
    }

    override fun finalize(): O = out

    private fun appenln() {
        if (prettyPrint && !ln) {
            out.append("\n")
            ln = true
        }
    }

    private fun indent() {
        if (prettyPrint) {
            if (!ln) {
                out.append("\n")
            }
            for (l in 0..level - 1) {
                out.append("  ")
            }
            ln = false
        }
    }
}

public fun <O : Appendable> O.appendHTML(prettyPrint : Boolean = true) : TagConsumer<O> = HTMLStreamBuilder(this, prettyPrint).delayed()

private val escapeMap = mapOf(
        '<' to "&lt;",
        '>' to "&gt;",
        '&' to "&amp;",
        '\'' to "&apos;",
        '\"' to "&quot;"
)

private fun Char._isLetter() = this in 'a' .. 'z' || this in 'A' .. 'Z'
private fun Char._isDigit() = this in '0' .. '9'
//private fun String.contains(ch : Char) = this.indexOf(ch) != -1

private fun String.isValidXmlAttributeName() =
        !this.toLowerCase().startsWith("xml")
                && this.isNotEmpty()
                && (this[0]._isLetter() || this[0] == '_')
                && this.all { it._isLetter() || it._isDigit() || "._:".indexOf(it) != -1 }

private fun Appendable.escapeAppend(s : CharSequence) {
    var lastIndex = 0
    for (idx in 0 .. s.length() - 1) {
        val ch = s[idx]
        val escape = escapeMap[ch]
        if (escape != null) {
            append(s, lastIndex, idx)
            append(escape)
            lastIndex = idx + 1
        }
    }

    if (lastIndex < s.length()) {
        append(s, lastIndex, s.length())
    }
}